/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package org.imca_cat.procrun;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.io.ByteArrayInputStream;
import java.util.Arrays;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;

import org.imca_cat.procrun.FixedSizeBufferInputStreamConsumer;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

/**
 * Tests {@link FixedSizeBufferInputStreamConsumer}.
 */
public class FixedSizeBufferInputStreamConsumerTest extends ProcrunTest {
  private ExecutorService executor;

  @Before
  public void setUp() {
    executor = Executors.newCachedThreadPool();
  }

  @After
  public void tearDown() {
    if (executor != null) executor.shutdownNow();
  }

  @Test
  public void testNoInput() throws ExecutionException, InterruptedException {
    ByteArrayInputStream stream = new ByteArrayInputStream(new byte[0]);
    FixedSizeBufferInputStreamConsumer consumer =
        new FixedSizeBufferInputStreamConsumer(stream, 128);
    byte[] expected = new byte[0];
    byte[] actual = byteBufferToArray(executor.submit(consumer).get());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testWithInput() throws ExecutionException, InterruptedException {
    byte[] expected = "abc".getBytes();
    ByteArrayInputStream stream = new ByteArrayInputStream(expected);
    FixedSizeBufferInputStreamConsumer consumer =
        new FixedSizeBufferInputStreamConsumer(stream, 128);
    byte[] actual = byteBufferToArray(executor.submit(consumer).get());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputSizeEqualToBufferSize() throws ExecutionException, InterruptedException {
    byte[] input = "abc".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer =
        new FixedSizeBufferInputStreamConsumer(stream, input.length);
    byte[] expected = input;
    byte[] actual = byteBufferToArray(executor.submit(consumer).get());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputBiggerThanBuffer01() throws ExecutionException, InterruptedException {
    byte[] input = "abcd".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(stream, 3);
    byte[] expected = "bcd".getBytes(US_ASCII);
    byte[] actual = byteBufferToArray(executor.submit(consumer).get());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputBiggerThanBuffer02() throws ExecutionException, InterruptedException {
    byte[] input = "abcde".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(stream, 3);
    byte[] expected = "cde".getBytes(US_ASCII);
    byte[] actual = byteBufferToArray(executor.submit(consumer).get());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputBiggerThanBuffer03() throws ExecutionException, InterruptedException {
    byte[] input = "abcdef".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(stream, 3);
    byte[] expected = "def".getBytes(US_ASCII);
    byte[] actual = byteBufferToArray(executor.submit(consumer).get());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputBiggerThanBuffer04() throws ExecutionException, InterruptedException {
    byte[] input = "abcdefg".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(stream, 3);
    byte[] expected = "efg".getBytes(US_ASCII);
    byte[] actual = byteBufferToArray(executor.submit(consumer).get());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputBiggerThanBuffer05() throws ExecutionException, InterruptedException {
    byte[] input = "ab".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(stream, 1);
    byte[] expected = "b".getBytes(US_ASCII);
    byte[] actual = byteBufferToArray(executor.submit(consumer).get());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputBiggerThanBuffer06() throws ExecutionException, InterruptedException {
    byte[] input = "abc".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(stream, 1);
    byte[] expected = "c".getBytes(US_ASCII);
    byte[] actual = byteBufferToArray(executor.submit(consumer).get());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputBiggerThanBuffer07() throws ExecutionException, InterruptedException {
    byte[] input = "abcdefghijklmnop".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer = new FixedSizeBufferInputStreamConsumer(stream, 6);
    byte[] expected = "klmnop".getBytes(US_ASCII);
    byte[] actual = byteBufferToArray(executor.submit(consumer).get());
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testInputBiggerThanBuffer08() throws ExecutionException, InterruptedException {
    byte[] input = "These aren't the droids you're looking for.".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer =
        new FixedSizeBufferInputStreamConsumer(stream, 19);
    String expected = "you're looking for.";
    String actual = new String(byteBufferToArray(executor.submit(consumer).get()), US_ASCII);
    assertEquals(expected, actual);
  }

  @Test
  public void testInputBiggerThanBuffer09() throws ExecutionException, InterruptedException {
    byte[] input = "These aren't the droids you're looking for.".getBytes(US_ASCII);
    ByteArrayInputStream stream = new ByteArrayInputStream(input);
    FixedSizeBufferInputStreamConsumer consumer =
        new FixedSizeBufferInputStreamConsumer(stream, 23);
    String expected = "ids you're looking for.";
    String actual = new String(byteBufferToArray(executor.submit(consumer).get()), US_ASCII);
    assertEquals(expected, actual);
  }
}
