/*
 * Copyright (c) 2013 J. Lewis Muir <jlmuir@imca-cat.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package org.imca_cat.procrun;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertNotEquals;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import java.io.IOException;
import java.io.InputStream;
import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.TimeoutException;

import org.imca_cat.procrun.InputStreamConsumer;
import org.imca_cat.procrun.InputStreamConsumerFactory;
import org.imca_cat.procrun.ProcessRunner;
import org.imca_cat.procrun.ProcessRunnerResult;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

/**
 * Tests {@link ProcessRunner}.
 */
public class ProcessRunnerTest extends ProcrunTest {
  private ExecutorService executor;

  @Before
  public void setUp() {
    executor = Executors.newCachedThreadPool();
  }

  @After
  public void tearDown() {
    if (executor != null) executor.shutdownNow();
  }

  @Test
  public void testExitStatus_zero() throws InterruptedException {
    ProcessRunnerResult result = ProcessRunner.run(new ProcessBuilder("true"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
  }

  @Test
  public void testExitStatus_nonzero() throws InterruptedException {
    ProcessRunnerResult result = ProcessRunner.run(new ProcessBuilder("false"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertNotEquals(0, result.exitStatus().intValue());
  }

  @Test
  public void testExitStatus_two() throws InterruptedException {
    ProcessRunnerResult result = ProcessRunner.run(new ProcessBuilder("sh", "-c", "exit 2"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(2, result.exitStatus().intValue());
  }

  @Test
  public void testStdout01() throws InterruptedException {
    ProcessRunnerResult result = ProcessRunner.run(new ProcessBuilder("printf", "abc"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    byte[] actual = byteBufferToArray(result.stdout());
    byte[] expected = "abc".getBytes();
    assertTrue(Arrays.equals(expected, actual));
    assertEquals("abc", result.stdoutString());
    assertTrue(result.hasStderr());
    assertEquals(0, result.stderr().remaining());
  }

  @Test
  public void testStdout02() throws InterruptedException {
    ProcessRunnerResult result =
        ProcessRunner.run(new ProcessBuilder("sh", "-c", "sleep 1; exec printf abc"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    byte[] actual = byteBufferToArray(result.stdout());
    byte[] expected = "abc".getBytes();
    assertTrue(Arrays.equals(expected, actual));
    assertEquals("abc", result.stdoutString());
    assertTrue(result.hasStderr());
    assertEquals(0, result.stderr().remaining());
  }

  @Test
  public void testStderr01() throws InterruptedException {
    ProcessRunnerResult result =
        ProcessRunner.run(new ProcessBuilder("sh", "-c", "exec 1>&2; exec printf abc"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    assertEquals(0, result.stdout().remaining());
    byte[] actual = byteBufferToArray(result.stderr());
    byte[] expected = "abc".getBytes();
    assertTrue(Arrays.equals(expected, actual));
    assertEquals("abc", result.stderrString());
  }

  @Test
  public void testStderr02() throws InterruptedException {
    ProcessRunnerResult result =
        ProcessRunner.run(new ProcessBuilder("sh", "-c", "sleep 1; exec 1>&2; exec printf abc"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    assertEquals(0, result.stdout().remaining());
    byte[] actual = byteBufferToArray(result.stderr());
    byte[] expected = "abc".getBytes();
    assertTrue(Arrays.equals(expected, actual));
    assertEquals("abc", result.stderrString());
  }

  @Test
  public void testStderr03() throws InterruptedException {
    ProcessRunnerResult result =
        ProcessRunner.run(new ProcessBuilder("sh", "-c",
            "printf 'error: bad motivator\\n' 1>&2; exit 1"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(1, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    assertEquals("", result.stdoutString(US_ASCII));
    assertTrue(result.hasStderr());
    assertEquals("error: bad motivator\n", result.stderrString(US_ASCII));
  }

  @Test
  public void testStdoutAndStderr01() throws InterruptedException {
    ProcessRunnerResult result =
        ProcessRunner.run(new ProcessBuilder("sh", "-c", "printf abc; exec 1>&2; exec printf def"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    byte[] actual;
    byte[] expected;
    actual = byteBufferToArray(result.stdout());
    expected = "abc".getBytes();
    assertTrue(Arrays.equals(expected, actual));
    actual = byteBufferToArray(result.stderr());
    expected = "def".getBytes();
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testStdoutAndStderr02() throws InterruptedException {
    ProcessRunnerResult result =
        ProcessRunner.run(new ProcessBuilder("sh", "-c", "printf abc 1>&2; exec printf def"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    byte[] actual;
    byte[] expected;
    actual = byteBufferToArray(result.stdout());
    expected = "def".getBytes();
    assertTrue(Arrays.equals(expected, actual));
    actual = byteBufferToArray(result.stderr());
    expected = "abc".getBytes();
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testStdoutAndStderr03() throws InterruptedException {
    ProcessRunnerResult result =
        ProcessRunner.run(new ProcessBuilder("sh", "-c",
            "printf abc; printf def 1>&2; exec printf ghi"));
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    byte[] actual;
    byte[] expected;
    actual = byteBufferToArray(result.stdout());
    expected = "abcghi".getBytes();
    assertTrue(Arrays.equals(expected, actual));
    actual = byteBufferToArray(result.stderr());
    expected = "def".getBytes();
    assertTrue(Arrays.equals(expected, actual));
  }

  @Test
  public void testStdoutConsumerException() throws InterruptedException {
    final IOException expected = new IOException("Bad motivator");
    InputStreamConsumerFactory f = new InputStreamConsumerFactory() {
      @Override
      public InputStreamConsumer newConsumer(InputStream s) {
        return new InputStreamConsumer() {
          @Override
          public ByteBuffer call() throws Exception {
            throw expected;
          };

          @Override
          public void cancel() {
          };
        };
      }
    };
    ProcessRunnerResult result = ProcessRunner.run(new ProcessBuilder("sleep", "5"), f, null);
    assertTrue(result.hasException());
    assertEquals(expected, result.exception());
  }

  @Test
  public void testStderrConsumerException() throws InterruptedException {
    final IOException expected = new IOException("Bad motivator");
    InputStreamConsumerFactory f = new InputStreamConsumerFactory() {
      @Override
      public InputStreamConsumer newConsumer(InputStream s) {
        return new InputStreamConsumer() {
          @Override
          public ByteBuffer call() throws Exception {
            throw expected;
          };

          @Override
          public void cancel() {
          };
        };
      }
    };
    ProcessRunnerResult result = ProcessRunner.run(new ProcessBuilder("sleep", "5"), null, f);
    assertTrue(result.hasException());
    assertEquals(expected, result.exception());
  }

  @Test
  public void testStdoutAndStderrConsumerException() throws InterruptedException {
    final IOException stdoutException = new IOException("Bad motivator: stdout");
    final IOException stderrException = new IOException("Bad motivator: stderr");
    InputStreamConsumerFactory fo = new InputStreamConsumerFactory() {
      @Override
      public InputStreamConsumer newConsumer(InputStream s) {
        return new InputStreamConsumer() {
          @Override
          public ByteBuffer call() throws Exception {
            throw stdoutException;
          };

          @Override
          public void cancel() {
          };
        };
      }
    };
    InputStreamConsumerFactory fe = new InputStreamConsumerFactory() {
      @Override
      public InputStreamConsumer newConsumer(InputStream s) {
        return new InputStreamConsumer() {
          @Override
          public ByteBuffer call() throws Exception {
            throw stderrException;
          };

          @Override
          public void cancel() {
          };
        };
      }
    };
    ProcessRunnerResult result = ProcessRunner.run(new ProcessBuilder("sleep", "5"), fo, fe);
    assertTrue(result.hasException());
    assertTrue(stdoutException == result.exception() || stderrException == result.exception());
  }

  @Test
  public void testInterrupt_noChild() throws CancellationException, ExecutionException,
      InterruptedException, TimeoutException {
    ProcessRunner runner = new ProcessRunner(new ProcessBuilder("sleep", "10"));
    Future<ProcessRunnerResult> f = executor.submit(runner);
    Thread.sleep(1000);
    executor.shutdownNow();
    boolean terminated = executor.awaitTermination(2L, TimeUnit.SECONDS);
    assertTrue(terminated);
    ProcessRunnerResult result = f.get(2L, TimeUnit.SECONDS);
    assertTrue(result.hasException());
    assertEquals(InterruptedException.class, result.exception().getClass());
  }

  @Test
  public void testInterrupt_childAttached() throws CancellationException, ExecutionException,
      InterruptedException, TimeoutException {
    ProcessRunner runner = new ProcessRunner(new ProcessBuilder("sh", "-c", "true && sleep 10"));
    Future<ProcessRunnerResult> f = executor.submit(runner);
    Thread.sleep(1000);
    executor.shutdownNow();
    boolean terminated = executor.awaitTermination(2L, TimeUnit.SECONDS);
    assertTrue(terminated);
    ProcessRunnerResult result = f.get(2L, TimeUnit.SECONDS);
    assertTrue(result.hasException());
    assertEquals(InterruptedException.class, result.exception().getClass());
  }

  @Test
  public void testInterrupt_childDetached() throws InterruptedException {
    ProcessRunner runner =
        new ProcessRunner(new ProcessBuilder("sh", "-c", "sleep 10 < /dev/null > /dev/null 2>&1"));
    executor.submit(runner);
    Thread.sleep(1000);
    executor.shutdownNow();
    boolean terminated = executor.awaitTermination(2L, TimeUnit.SECONDS);
    assertTrue(terminated);
  }

  @Test
  public void testInterrupt_immediately() throws CancellationException, ExecutionException,
      InterruptedException, TimeoutException {
    ProcessRunner runner = new ProcessRunner(new ProcessBuilder("sleep", "10"));
    Future<ProcessRunnerResult> f = executor.submit(runner);
    executor.shutdownNow();
    boolean terminated = executor.awaitTermination(2L, TimeUnit.SECONDS);
    assertTrue(terminated);
    f.get(2L, TimeUnit.SECONDS);
  }

  @Test
  public void testInterrupt_immediatelyWithChild() throws CancellationException,
      ExecutionException, InterruptedException, TimeoutException {
    ProcessRunner runner = new ProcessRunner(new ProcessBuilder("sh", "-c", "true && sleep 10"));
    Future<ProcessRunnerResult> f = executor.submit(runner);
    executor.shutdownNow();
    boolean terminated = executor.awaitTermination(2L, TimeUnit.SECONDS);
    assertTrue(terminated);
    f.get(2L, TimeUnit.SECONDS);
  }

  @Test
  public void testCancel_noChild() throws CancellationException, ExecutionException,
      InterruptedException, TimeoutException {
    ProcessRunner runner = new ProcessRunner(new ProcessBuilder("sleep", "10"));
    Future<ProcessRunnerResult> f = executor.submit(runner);
    Thread.sleep(1000);
    f.cancel(true);
    try {
      f.get(2L, TimeUnit.SECONDS);
      fail();
    } catch (CancellationException e) {
      /* EXPECTED */
    }
    executor.shutdown();
    boolean terminated = executor.awaitTermination(2L, TimeUnit.SECONDS);
    assertTrue(terminated);
  }

  @Test
  public void testCancel_childAttached() throws CancellationException, ExecutionException,
      InterruptedException, TimeoutException {
    ProcessRunner runner = new ProcessRunner(new ProcessBuilder("sh", "-c", "true && sleep 10"));
    Future<ProcessRunnerResult> f = executor.submit(runner);
    Thread.sleep(1000);
    f.cancel(true);
    try {
      f.get(2L, TimeUnit.SECONDS);
      fail();
    } catch (CancellationException e) {
      /* EXPECTED */
    }
    executor.shutdown();
    boolean terminated = executor.awaitTermination(2L, TimeUnit.SECONDS);
    assertTrue(terminated);
  }

  @Test
  public void testCancel_immediately() throws CancellationException, ExecutionException,
      InterruptedException, TimeoutException {
    ProcessRunner runner = new ProcessRunner(new ProcessBuilder("sleep", "10"));
    Future<ProcessRunnerResult> f = executor.submit(runner);
    boolean successfulCancel = f.cancel(true);
    assertTrue(successfulCancel);
  }

  @Test
  public void testCancel_immediatelyWithChild() throws CancellationException, ExecutionException,
      InterruptedException, TimeoutException {
    ProcessRunner runner = new ProcessRunner(new ProcessBuilder("sh", "-c", "true && sleep 10"));
    Future<ProcessRunnerResult> f = executor.submit(runner);
    boolean successfulCancel = f.cancel(true);
    assertTrue(successfulCancel);
  }

  @Test
  public void testRedirectedStderr() throws InterruptedException {
    ProcessBuilder builder =
        new ProcessBuilder("sh", "-c", "printf abc; sleep 1; exec 1>&2; exec printf def");
    builder.redirectErrorStream(true);
    ProcessRunnerResult result = ProcessRunner.run(builder);
    assertFalse(result.hasException());
    assertTrue(result.hasExitStatus());
    assertEquals(0, result.exitStatus().intValue());
    assertTrue(result.hasStdout());
    byte[] actual = byteBufferToArray(result.stdout());
    byte[] expected = "abcdef".getBytes();
    assertTrue(Arrays.equals(expected, actual));
    assertEquals("abcdef", result.stdoutString());
    assertTrue(result.hasStderr());
    assertEquals("", result.stderrString());
  }
}
